#include</build/j4-dmenu-desktop/src/build/generated/version.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/main.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/AppManager.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/Application.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/CMDLineAssembler.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/CMDLineTerm.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/Dmenu.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/FieldCodes.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/FileFinder.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/Formatters.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/HistoryManager.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/I3Exec.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/LineReader.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/LocaleSuffixes.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/SearchPath.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/Utilities.cc>
#include</build/j4-dmenu-desktop/src/build/../j4-dmenu-desktop-r3.2/src/NotifyInotify.cc>
